/*
 * Control interface to generic front ends.
 * written/copyrights 1997/99 by Andreas Neuhaus (and Michael Hipp)
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>       /* strrchr()                    */
#include <math.h>         /* floor()                      */

#include <sys/socket.h>

#include "callbacks.h"    /* audio_reopen_arg_t           */

#include "ogg123.h"
#include "buffer.h"
#include "i18n.h"
#include "status.h"

#include "control_generic.h"

#define ONLY_DISPLAY_STREAM_INFO_ON_INIT 0

#define MODE_STOPPED 0
#define MODE_PLAYING 1
#define MODE_PAUSED 2

extern ogg123_options_t options;
double seconds_per_chunk = 0.0;

static char *createmsg(const char *fmt, ...)
{
  static char buffer[1024]; // big enough for a message? (i think so).
  int size = 0;
  va_list ap;

  va_start(ap, fmt);
  size += vsnprintf(buffer, 1024 - 2, fmt, ap);
  va_end(ap);

  buffer[size++] = '\0';

  return strdup(buffer);
}


/*
// calculate statistics and either print them directly or schedule an
// action (buffer mode).
//
*/
static void generic_sendstat(buf_t *audio_buffer, decoder_t *decoder,
                             int chunk_number)
{
  decoder_stats_t *decoder_statistics;
  int samples_per_chunk;
  int total_chunks;
  char *message;
  int remaining_chunks;
  double remaining_time;

  decoder_statistics  = decoder->format->statistics(decoder);
  samples_per_chunk   = decoder_statistics->current_pcm_sample /
                        chunk_number;
  total_chunks  = ceil(decoder_statistics->total_pcm_samples /
                       (double)samples_per_chunk);

  remaining_chunks  = total_chunks - chunk_number;
  remaining_time    = decoder_statistics->total_time - decoder_statistics->current_time;

  if (remaining_time < 0) {
    // happens on a non-seekable stream (e.g. stream over http)
    //
    remaining_chunks  = 0;
    remaining_time    = 0.0;
  }
  message = createmsg(
              "@F %d %d %3.2f %3.2f %ld",
              chunk_number,
              remaining_chunks,
              decoder_statistics->current_time,
              remaining_time,
              decoder_statistics->instant_bitrate);
  if (audio_buffer) {
    /* Place a status update into the buffer */
    buffer_append_action_at_end(audio_buffer,
                                &remote_write_action,
                                message);
  } else {
    remote_write_action(audio_buffer, message);
  }

  /*
  // This dumb-down is needed to mimic the frame-based JUMP
  // functionality of mpg123. Ogg-Vorbis can jump to a time but mpg123
  // does it per frame (yep, sucks).
  //
  // This is used in control_generic() when handling the JUMP command.
  //
  */
  if (!seconds_per_chunk) {
    seconds_per_chunk  = decoder_statistics->total_time / total_chunks;
  }

  free(decoder_statistics);
}

static void stop_play(format_t *format, decoder_t *decoder,
                      transport_t *transport, data_source_t *source,
                      buf_t *audio_buffer, int instant)
{
  if (audio_buffer) {
    if (!instant) {
      buffer_wait_for_empty(audio_buffer);
    }
    buffer_thread_kill(audio_buffer);
  }

  if (format && decoder) {
    format->cleanup(decoder);
  }
  if (transport && source) {
    transport->close(source);
  }
}

void control_generic(audio_play_arg_t *audio_play_arg, buf_t *audio_buffer)
{
  struct timeval tv;
  fd_set fds;
  int init = 0;
  int mode = MODE_STOPPED;
  int n;
  int ret;
  int chunk_no;  /* in mpg123 this is called a "frame" */
  audio_format_t old_audio_fmt = { 0, 0, 0, 0, 0 };
  audio_format_t new_audio_fmt;
  audio_reopen_arg_t *reopen_arg;
  decoder_t *decoder = NULL;
  decoder_callbacks_t decoder_callbacks;
  void *decoder_callbacks_arg;
  transport_t *transport = NULL;
  format_t *format = NULL;
  data_source_t *source = NULL;
  int convsize = AUDIO_CHUNK_SIZE;
  char *convbuffer;
  int eos;
  char *source_string;
  int stream_change = 0;

  convbuffer = malloc(convsize);
  if (convbuffer == NULL) {
    status_error(_("Error: Out of memory in control_generic().\n"));
    exit(1);
  }

  /* Set preferred audio format (used by decoder) */
  new_audio_fmt.big_endian = ao_is_big_endian();
  new_audio_fmt.signed_sample = 1;
  new_audio_fmt.word_size = 2;

  /* Select appropriate callbacks */
  if (audio_buffer) {
    decoder_callbacks.printf_error = &decoder_buffered_error_callback;
    decoder_callbacks_arg = audio_buffer;
  } else {
    decoder_callbacks.printf_error = &decoder_error_callback;
    decoder_callbacks_arg = NULL;
  }
  decoder_callbacks.printf_metadata = NULL;

  // Disable buffering on stdout. Otherwise playmp3list gets really jerky.
  setvbuf(stdout, (char *)NULL, _IOLBF, 0);

  remote_write_action(NULL, createmsg("@R OGG123 %s", VERSION)); /* could also print PACKAGE */

  while (1) {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    FD_ZERO(&fds);
    FD_SET(STDIN_FILENO, &fds);

    /* play frame if no command needs to be processed */
    if (mode == MODE_PLAYING) {
      n = select(STDIN_FILENO+1, &fds, NULL, NULL, &tv);
      if (n == 0) {
        /* Read another block of audio data */
        ret = format->read(decoder, convbuffer, convsize, &eos, &new_audio_fmt);

        if (ret == 0) {
          eos = 1;
          mode = MODE_STOPPED;
          if (audio_buffer) {
            buffer_append_action_at_end(audio_buffer,
                                        &remote_write_action,
                                        createmsg("@P 0 EOF"));
            buffer_mark_eos(audio_buffer);
          } else {
            remote_write_action_const(audio_buffer, "@P 0 EOF");
          }
          continue;
        } else if (ret < 0) {
          status_error(_("Error: Decoding failure.\n"));
          continue;
        }

        /* Check to see if the audio format has changed */
        if (!audio_format_equal(&new_audio_fmt, &old_audio_fmt)) {
          old_audio_fmt = new_audio_fmt;

          reopen_arg = new_audio_reopen_arg(options.devices, &new_audio_fmt);

          if (audio_buffer) {
            buffer_insert_action_at_end(audio_buffer, &audio_reopen_action,
                                        reopen_arg);
          } else {
            audio_reopen_action(NULL, reopen_arg);
          }

          stream_change = 1;
        }

        if (init) {
#if UGLY_ID3_HACK
          char *id3_string = format->id3string(decoder);
          char *message;

          // if there is no title (first field) we use the filename
          //
          if (id3_string[0] != ' ') {
            message = createmsg("@I ID3:%s", id3_string);
            remote_write_action(NULL, message);
          } else {
#else
            generic_sendinfo(source_string);
#endif
#if UGLY_ID3_HACK
          }
          free(id3_string);
#endif
        }

        if (stream_change || init) {
          char *message;
          decoder_stats_t *decoder_statistics;
          const char *mode_string;

          /* static const char *modes[4] = {"Stereo", "Joint-Stereo", "Dual-Channel", "Single-Channel"}; */
          /* mode (read vorbis/doc/stereo.html) i don't see it in the file format */
          switch(new_audio_fmt.channels) {
            case 1:
              mode_string = "Single-Channel";
              break;
            case 2:
              mode_string = "Stereo";
              break;
            default:
              mode_string = "UNDEFINED";
          }

          decoder_statistics = decoder->format->statistics(decoder);
          message = createmsg("@S %s %d %ld %s %d %d %d %d %d %d %d %d",
                              "0.0", /* mpeg(!) type FIXME */
                              0, /* layer FIXME */
                              new_audio_fmt.rate, /* sampling frequency */
                              mode_string,
                              0, /* mode extension */
                              AUDIO_CHUNK_SIZE,
                              new_audio_fmt.channels,  /* stereo */
                              0, /* copyright (1|0) FIXME (is in vorbis comments) */
                              1, /* error protection (ogg is designed to provide this, is this optional?) */
                              0, /* emphasis, obsolete piece of mp3 header */
                              decoder_statistics->avg_bitrate / 1024, /* tabsel_123[fr->lsf][fr->lay-1][fr->bitrate_index], */
                              0);/* fr->extension); */
          remote_write_action(NULL, message);
          free(decoder_statistics);
          stream_change = 0;
        }

        init = 0;
        chunk_no++;
        generic_sendstat(audio_buffer, decoder, chunk_no);

        if (audio_buffer) {
          buffer_submit_data(audio_buffer, convbuffer, ret);
        } else {
          audio_play_callback(convbuffer, ret, eos, audio_play_arg);
        }
      }
    } else {
      /* wait for command */
      while (1) {
        tv.tv_sec  = 1;
        tv.tv_usec = 0;
        FD_ZERO(&fds);
        FD_SET(STDIN_FILENO, &fds);
        n = select(STDIN_FILENO+1, &fds, NULL, NULL, &tv);
        if (n != 0) {
          break;
        }
      }
    }
    /* exit on error */
    if (n < 0) {
      fprintf(stderr, "Error waiting for command: %s\n", strerror(errno));
      exit(1);
    }

    /* process command */
    if (n > 0) {
      int len,rc;
      char buf[1024];
      char *cmd;

      /* read command */
      buf[0] = 0; len = 0;
      do {
        rc = read(STDIN_FILENO, buf+len, 1);
        if (rc > 0) {
          len += rc;
        }
      } while (rc>0 && buf[len-rc]!='\n' && buf[len-rc]!='\r');

      buf[len] = 0;

      /* exit on error */
      if (rc < 0) {
        fprintf(stderr, "Error reading command: %s\n", strerror(errno));
        exit(1);
      }

      if (len == 0) {
        fprintf(stderr, "Caught End-Of-File -> Exit\n");
        exit(1);
      }

      /* strip CR/LF at EOL */
      while (len>0 && (buf[strlen(buf)-1] == '\n' || buf[strlen(buf)-1] == '\r')) {
        buf[strlen(buf)-1] = 0;
        len--;
      }

      /* continue if no command */
      if (len == 0) {
        continue;
      }
      cmd = strtok(buf, " \t");
      if (!cmd || !strlen(cmd)) {
        continue;
      }

      /*
      // The commands
      //
      */
      //fprintf(stderr,"CMD: %s\n",cmd);

      /* QUIT */
      if (!strcasecmp(cmd, "Q") || !strcasecmp(cmd, "QUIT")) {
        break;
      }

      /* LOAD */
      if (!strcasecmp(cmd, "L") || !strcasecmp(cmd, "LOAD")) {
        source_string = strtok(NULL, "");
        if (mode != MODE_STOPPED) {
          stop_play(format, decoder, transport, source, audio_buffer, 1);
          mode = MODE_STOPPED;
        }

        /* Locate and use transport for this data source */
        if ((transport = select_transport(source_string)) == NULL) {
          status_error(_("No module could be found to read from %s.\n"), source_string);
          continue;
        }
        if ((source = transport->open(source_string, &options, audio_play_arg->stat_format)) == NULL) {
          status_error(_("Cannot open %s.\n"), source_string);
          continue;
        }
        /* Detect the file format and initialize a decoder */
        if ((format = select_format(source)) == NULL) {
          status_error(_("The file format of %s is not supported.\n"), source_string);
          continue;
        }
        if ((decoder = format->init(source, &options, &new_audio_fmt,
                                     &decoder_callbacks,
                                     decoder_callbacks_arg)) == NULL) {
          status_error(_("Error opening %s using the %s module."
                         "  The file may be corrupted.\n"), source_string,
                       format->name);
          continue;
        }

        /* Start the audio playback thread before we begin sending data */
        if (audio_buffer) {
          /* First reset mutexes and other synchronization variables */
          buffer_reset(audio_buffer);
          buffer_thread_start(audio_buffer);
        }

        mode = MODE_PLAYING;
        init = 1;
        chunk_no = 0;
        eos = 0;
        continue;
      }

      /* STOP */
      if (!strcasecmp(cmd, "S") || !strcasecmp(cmd, "STOP")) {
        if (mode != MODE_STOPPED) {
          stop_play(format, decoder, transport, source, audio_buffer, 1);
          mode = MODE_STOPPED;
          remote_write_action_const(NULL, "@P 0");
        }
        continue;
      }

      /* PAUSE */
      if (!strcasecmp(cmd, "P") || !strcasecmp(cmd, "PAUSE")) {
        if (mode == MODE_STOPPED) {
          continue;
        }
        if (mode == MODE_PLAYING) {
          mode = MODE_PAUSED;
          if (audio_buffer) {
            buffer_thread_pause (audio_buffer);
          }
          remote_write_action_const(NULL, "@P 1");
        } else {
          mode = MODE_PLAYING;
          if (audio_buffer) {
            buffer_thread_unpause (audio_buffer);
          }
          remote_write_action_const(NULL, "@P 2");
        }
        continue;
      }

      /* JUMP */
      /*
      // This mimics the mpg123 frame behaviour. This makes the chunk_no
      // go wrong (and the total_chunks because of this).
      //
      // Maybe there is a fix possible but i don't see it yet. All the floating
      // point magic doesn't help either.
      //
      */
      if (!strcasecmp(cmd, "J") || !strcasecmp(cmd, "JUMP")) {
        char *spos;
        int seek_frame;
        double seek;
        int whence;

        spos = strtok(NULL, " \t");
        if (!spos) {
          continue;
        }

        /*
        // After this chunk_no _could_ be wrong. I don't think this hurts
        // but we'll see.
        //
        */
        if (spos[0] == '-') {
          seek_frame = atoi(spos);
          whence = DECODER_SEEK_CUR;
          chunk_no += (seek_frame);
        } else if (spos[0] == '+') {
          seek_frame = atoi(spos+1);
          whence = DECODER_SEEK_CUR;
          chunk_no += (seek_frame - 1);
        } else {
          seek_frame = atoi(spos);
          whence = DECODER_SEEK_START;
          chunk_no = (seek_frame - 1);
        }
        seek = seek_frame * seconds_per_chunk;

        if (mode == MODE_STOPPED) {
          continue;
        }

        /*
        // Doesn't work nor nonseekable streams. This could be fixed with
        // doing forward reading. Only forward and jumping to positions
        // further in the stream would work this way, but it's better than
        // nothing.
        //
        */
        format->seek(decoder, seek, whence);
        continue;
      }

      /* EQUALIZE Band-0-Left Band-0-Right Band-1-Left Band-1-Right ... */
      if (!strcasecmp(cmd, "E") || !strcasecmp(cmd, "EQUALIZE")) {
        // NOT IMPLEMENTED, no need to print an error though.
        continue;
      }

      /* unknown command */
      remote_write_action(NULL, createmsg("@E Unknown command '%s'", cmd));
    }
  }

  if (mode != MODE_STOPPED) {
    if (mode == MODE_PAUSED) {
      // go for the instant kill, if we wait for the buffer we deadlock.
      stop_play(format, decoder, transport, source, audio_buffer, 1);
    } else {
      stop_play(format, decoder, transport, source, audio_buffer, 0);
    }
  }

  free(convbuffer);
}

/* EOF */
